from tests.util import *


class TestRadius(TestCase):

    def __init__(self, *args, **kwargs):
        mk_temp_dirs()
        self.using_vc = self.init_vc()
        super(TestRadius, self).__init__(*args, **kwargs)

    def setUp(self):
        success = self.init_vc()
        if not success:
            raise SkipTest("%s not configured correctly" % self.vc)

    def check(self, original, modified, expected,
              test_name='check', options=None,
              cwd=TEMP_DIR, apply_config=False):
        """Modify original to modified, and check that pep8radius turns this
        into expected."""
        temp_file = os.path.join(cwd, 'temp.py')

        options = parse_args(options, apply_config=apply_config)

        # TODO remove this color hack, and actually test printing color diff
        options.no_color = True

        with open(temp_file, 'w') as f:
            f.write(original)
        committed = self.successfully_commit_files([temp_file],
                                                   commit=test_name)

        with open(temp_file, 'w') as f:
            f.write(modified)

        options.verbose = 1
        r = Radius(vc=self.vc, options=options, cwd=cwd)
        with captured_output() as (out, err):
            r.fix()
        self.assertIn('would fix', out.getvalue())
        self.assertNotIn('@@', out.getvalue())
        options.verbose = 0

        options.diff = True
        r = Radius(vc=self.vc, options=options, cwd=cwd)
        with captured_output() as (out, err):
            r.fix()
        exp_diff = get_diff(modified, expected, temp_file)
        self.assert_equal(out.getvalue(), exp_diff, test_name)
        options.diff = False

        options.in_place = True
        r = Radius(vc=self.vc, options=options, cwd=cwd)
        # Run pep8radius
        r.fix()

        with open(temp_file, 'r') as f:
            result = f.read()
        self.assert_equal(result, expected, test_name)

        # Run pep8radius again, it *should* be that this doesn't do anything.
        out = pep8radius_main(options, vc=self.vc, cwd=cwd)
        self.assertEqual(out, '')

        with open(temp_file, 'r') as f:
            result = f.read()
        self.assert_equal(result, expected, test_name)

    def assert_equal(self, result, expected, test_name):
        """like assertEqual but with a nice diff output if not equal."""
        self.assertEqual(result, expected,
                         get_diff(expected, result, test_name,
                                  'expected', 'result'))


class MixinTests:

    """All Radius tests are placed in this class."""

    def test_one_line(self):
        original = 'def poor_indenting():\n  a = 1\n  b = 2\n  return a + b\n\nfoo = 1; bar = 2; print(foo * bar)\na=1; b=2; c=3\nd=7\n\ndef f(x = 1, y = 2):\n    return x + y\n'
        modified = 'def poor_indenting():\n  a = 1\n  b = 2\n  return a + b\n\nfoo = 1; bar = 2; print(foo * bar)\na=1; b=42; c=3\nd=7\n\ndef f(x = 1, y = 2):\n    return x + y\n'
        expected = 'def poor_indenting():\n  a = 1\n  b = 2\n  return a + b\n\nfoo = 1; bar = 2; print(foo * bar)\na = 1\nb = 42\nc = 3\nd=7\n\ndef f(x = 1, y = 2):\n    return x + y\n'
        self.check(original, modified, expected, 'test_one_line')

    def test_one_line_from_subdirectory(self):
        original = 'def poor_indenting():\n  a = 1\n  b = 2\n  return a + b\n\nfoo = 1; bar = 2; print(foo * bar)\na=1; b=2; c=3\nd=7\n\ndef f(x = 1, y = 2):\n    return x + y\n'
        modified = 'def poor_indenting():\n  a = 1\n  b = 2\n  return a + b\n\nfoo = 1; bar = 2; print(foo * bar)\na=1; b=42; c=3\nd=7\n\ndef f(x = 1, y = 2):\n    return x + y\n'
        expected = 'def poor_indenting():\n  a = 1\n  b = 2\n  return a + b\n\nfoo = 1; bar = 2; print(foo * bar)\na = 1\nb = 42\nc = 3\nd=7\n\ndef f(x = 1, y = 2):\n    return x + y\n'
        self.check(original, modified, expected, 'test_one_line',
                   cwd=SUBTEMP_DIR)

    def test_with_docformatter(self):
        original = 'def poor_indenting():\n  """       Great function"""\n  a = 1\n  b = 2\n  return a + b\n\n\n\nfoo = 1; bar = 2; print(foo * bar)\na=1; b=2; c=3\nd=7\n\ndef f(x = 1, y = 2):\n    return x + y\n'
        modified = 'def poor_indenting():\n  """  Very great function"""\n  a = 1\n  b = 2\n  return a + b\n\n\n\nfoo = 1; bar = 2; print(foo * bar)\na=1; b=42; c=3\nd=7\n\ndef f(x = 1, y = 2):\n    return x + y\n'
        expected = 'def poor_indenting():\n  """  Very great function"""\n  a = 1\n  b = 2\n  return a + b\n\n\n\nfoo = 1; bar = 2; print(foo * bar)\na = 1\nb = 42\nc = 3\nd=7\n\ndef f(x = 1, y = 2):\n    return x + y\n'
        self.check(original, modified, expected, 'test_without_docformatter')

        expected = 'def poor_indenting():\n  """Very great function."""\n  a = 1\n  b = 2\n  return a + b\n\n\n\nfoo = 1; bar = 2; print(foo * bar)\na = 1\nb = 42\nc = 3\nd=7\n\ndef f(x = 1, y = 2):\n    return x + y\n'
        self.check(original, modified, expected,
                   'test_with_docformatter', ['--docformatter'])

    def test_bad_rev(self):
        with captured_output() as (out, err):
            self.assertRaises(CalledProcessError,
                              lambda x: Radius(rev=x,
                                               vc=self.vc,
                                               cwd=TEMP_DIR),
                              'random_junk_sha')

    def test_earlier_revision(self):
        if self.vc == 'bzr':
            raise SkipTest("TODO get me working")

        start = self.save_and_commit('a=1;', 'AAA.py')
        self.checkout('ter', create=True)
        self.save_and_commit('b=1;', 'BBB.py')
        tip = self.save_and_commit('c=1;', 'CCC.py')
        save('c=1', 'CCC.py')

        args = parse_args(['--diff', '--no-color'])
        r = Radius(rev=start, options=args, vc=self.vc, cwd=TEMP_DIR)
        with captured_output() as (out, err):
            r.fix()
        diff = out.getvalue()

        files = [os.path.join(TEMP_DIR, f) for f in ['BBB.py', 'CCC.py']]

        exp_diff = get_diff_many(['b=1;', 'c=1'],
                                 ['b = 1\n', 'c = 1\n'],
                                 files)
        self.assert_equal(diff, exp_diff, 'earlier_revision')

        # TODO test the diff is correct

    def test_deleted_file(self):
        os.remove(os.path.join(TEMP_DIR, 'a.py'))
        args = parse_args(['--diff', '--no-color'])
        r = Radius(options=args, vc=self.vc, cwd=TEMP_DIR)
        with captured_output() as (out, err):
            r.fix()
        self.assertEqual(out.getvalue(), '')

    def test_exclude(self):
        self.save_and_commit('b=1;', 'BBB.py')
        save('b=1', 'BBB.py')
        args = parse_args(['--diff', '--no-color', '--exclude=BBB.py'])
        r = Radius(options=args, vc=self.vc, cwd=TEMP_DIR)
        with captured_output() as (out, err):
            r.fix()
        self.assertEqual(out.getvalue(), '')

    def test_config(self):
        cfg = os.path.join(TEMP_DIR, '.pep8')
        with open(cfg, mode='w') as f:
            f.write("[pep8]\nindent-size=2")
        original = "def f(x):\n    return 2*x\n"
        modified = "def f(x):\n    return 3*x\n"
        expected = "def f(x):\n  return 3 * x\n"
        self.check(original, modified, expected,
                   'test_config', ['--config=%s' % cfg],
                   apply_config=True)
        remove(cfg)

if __name__ == '__main__':
    test_main()



"""This module defines the version control specific functionality, we need.

The aim is to work, for example to get the root directory, independent
of which version control system we are using. That is, subclasses of the
abstract class VersionControl are specific to vcs e.g. Git.

"""


import os
import re

from pep8radius.shell import (shell_out, shell_out_ignore_exitcode,
                              CalledProcessError)  # with 2.6 compat


class AbstractMethodError(NotImplementedError):
    pass


def using_git(cwd):
    """Test whether the directory cwd is contained in a git repository."""
    try:
        git_log = shell_out(["git", "log"], cwd=cwd)
        return True
    except (CalledProcessError, OSError):  # pragma: no cover
        return False


def using_hg(cwd):
    """Test whether the directory cwd is contained in a mercurial
    repository."""
    try:
        hg_log = shell_out(["hg",   "log"], cwd=cwd)
        return True
    except (CalledProcessError, OSError):
        return False


def using_bzr(cwd):
    """Test whether the directory cwd is contained in a bazaar repository."""
    try:
        bzr_log = shell_out(["bzr", "log"], cwd=cwd)
        return True
    except (CalledProcessError, OSError):
        return False


class VersionControl(object):

    """Abstract base class for defining the methods we need to work with a
    version control system."""

    def __init__(self, cwd=None):
        self.root = os.path.abspath(self.root_dir(cwd=cwd))

    def _shell_out(self, *args, **kwargs):
        return shell_out(*args, cwd=self.root, **kwargs)

    @staticmethod
    def from_string(vc):
        """Return the VersionControl superclass from a string, for example
        VersionControl.from_string('git') will return Git."""
        try:
            # Note: this means all version controls must have
            # a title naming convention (!)
            vc = globals()[vc.title()]
            assert(issubclass(vc, VersionControl))
            return vc
        except (KeyError, AssertionError):
            raise NotImplementedError("Unknown version control system.")

    @staticmethod
    def which(cwd=None):  # pragma: no cover
        """Try to find which version control system contains the cwd directory.

        Returns the VersionControl superclass e.g. Git, if none were
        found this will raise a NotImplementedError.

        """
        if cwd is None:
            cwd = os.getcwd()
        for (k, using_vc) in globals().items():
            if k.startswith('using_') and using_vc(cwd=cwd):
                return VersionControl.from_string(k[6:])

        # Not supported (yet)
        raise NotImplementedError("Unknown version control system, "
                                  "or you're not in the project directory.")

    # abstract methods
    @staticmethod
    def file_diff_cmd(r, file_name):  # pragma: no cover
        raise AbstractMethodError()

    @staticmethod
    def filenames_diff_cmd(r):  # pragma: no cover
        raise AbstractMethodError()

    @staticmethod
    def parse_diff_filenames(diff_files):  # pragma: no cover
        raise AbstractMethodError()

    @staticmethod
    def root_dir(cwd=None):  # pragma: no cover
        raise AbstractMethodError()

    def current_branch(self):  # pragma: no cover
        raise AbstractMethodError()

    @staticmethod
    def merge_base(rev1, rev2):  # pragma: no cover
        raise AbstractMethodError()

    def branch_point(self, rev=None):
        current = self.current_branch()
        if rev is None:
            return current
        else:
            return self.merge_base(rev, current)

    def modified_lines(self, r, file_name):
        """Returns the line numbers of a file which have been changed."""
        cmd = self.file_diff_cmd(r, file_name)
        diff = shell_out_ignore_exitcode(cmd, cwd=self.root)
        return self.modified_lines_from_diff(diff)

    def modified_lines_from_diff(self, diff):
        """Returns the changed lines in a diff.

        - Potentially this is vc specific (if not using udiff).

        Note: this returns the line numbers in descending order.

        """
        from pep8radius.diff import modified_lines_from_udiff
        for start, end in modified_lines_from_udiff(diff):
            yield start, end

    def get_filenames_diff(self, r):
        """Get the py files which have been changed since rev."""
        cmd = self.filenames_diff_cmd(r)

        diff_files = shell_out_ignore_exitcode(cmd, cwd=self.root)
        diff_files = self.parse_diff_filenames(diff_files)

        py_files = set(f for f in diff_files if f.endswith('.py'))

        if r.options.exclude:
            # TODO do we have to take this from root dir?
            from glob import fnmatch
            for pattern in r.options.exclude:
                py_files.difference_update(fnmatch.filter(py_files, pattern))
        return sorted(os.path.join(self.root, file_name)
                      for file_name in py_files)


class Git(VersionControl):

    def current_branch(self):
        return self._shell_out(["git", "rev-parse", "HEAD"])

    @staticmethod
    def root_dir(cwd=None):
        root = shell_out(['git', 'rev-parse', '--show-toplevel'], cwd=cwd)
        return os.path.normpath(root)

    def merge_base(self, rev1, rev2):
        return self._shell_out(['git', 'merge-base', rev1, rev2])

    @staticmethod
    def file_diff_cmd(r, f):
        """Get diff for one file, f."""
        return ['git', 'diff', r.rev, f]

    @staticmethod
    def filenames_diff_cmd(r):
        """Get the names of the py files in diff."""
        return ['git', 'diff', r.rev, '--name-only']

    @staticmethod
    def parse_diff_filenames(diff_files):
        """Parse the output of filenames_diff_cmd."""
        return diff_files.splitlines()


class Hg(VersionControl):

    def current_branch(self):
        return self._shell_out(["hg", "id"])[:12]  # this feels awkward

    @staticmethod
    def root_dir(cwd=None):
        return shell_out(['hg', 'root'], cwd=cwd)

    def merge_base(self, rev1, rev2):
        output = self._shell_out(['hg', 'debugancestor', rev1, rev2])
        return output.split(':')[1]

    @staticmethod
    def file_diff_cmd(r, f):
        """Get diff for one file, f."""
        return ['hg', 'diff', '-r', r.rev, f]

    @staticmethod
    def filenames_diff_cmd(r):
        """Get the names of the py files in diff."""
        return ["hg", "diff", "--stat", "-r", r.rev]

    @staticmethod
    def parse_diff_filenames(diff_files):
        """Parse the output of filenames_diff_cmd."""
        # one issue is that occasionaly you get stdout from something else
        # specifically I found this in Coverage.py, luckily the format is
        # different (at least in this case)
        it = re.findall('(\n|^) ?(?P<file_name>.*\.py)\s+\|', diff_files)
        return [t[1] for t in it]


class Bzr(VersionControl):

    def current_branch(self):
        return self._shell_out(["bzr", "version-info",
                                "--custom", "--template={revision_id}"])

    @staticmethod
    def root_dir(cwd=None):
        return shell_out(['bzr', 'root'], cwd=cwd)

    def merge_base(self, rev1, rev2):
        # Note: find-merge-base just returns rev1 if rev2 is not found
        # we assume that rev2 is a legitamate revision.
        # the following raise a CalledProcessError if it's a bad revision
        shell_out(['bzr', 'log', '-c', rev1], cwd=self.root)

        output = shell_out_ignore_exitcode(['bzr', 'find-merge-base',
                                            rev1, rev2],
                                           cwd=self.root)
        # 'merge base is revision name@example.com-20140602232408-d3wspoer3m35'
        return output.rsplit(' ', 1)[1]

    @staticmethod
    def file_diff_cmd(r, f):
        """Get diff for one file, f."""
        return ['bzr', 'diff', f, '-r', r.rev]

    @staticmethod
    def filenames_diff_cmd(r):
        """Get the names of the py files in diff."""
        # TODO Can we do this better (without parsing the entire diff?)
        return ['bzr', 'status', '-S', '-r', r.rev]  # TODO '--from-root' ?

    @staticmethod
    def parse_diff_filenames(diff_files):
        """Parse the output of filenames_diff_cmd."""
        # ?   .gitignore
        # M  0.txt
        files = []
        for line in diff_files.splitlines():
            line = line.strip()
            fn = re.findall('[^ ]+\s+(.*.py)', line)
            if fn and not line.startswith('?'):
                files.append(fn[0])
        return files



from tests.util import *


class TestUDiffParsing(TestCase):

    def test_udiff_parsing(self):
        with open(os.path.join(TEST_DIR, 'diff1.txt')) as f:
            example_udiff = f.read()
        lines = list(modified_lines_from_udiff(example_udiff))
        assert(lines == [(444, 444), (424, 429), (54, 56)])


if __name__ == '__main__':
    test_main()



"""This module contains the helpers for diff creation, printing and extracting
line numbers."""

from __future__ import print_function

import os
import re


def modified_lines_from_udiff(udiff):
    """Extract from a udiff an iterator of tuples of (start, end) line numbers.

    Note: these are returned in descending order.

    """
    chunks = re.split('\n@@ [^\n]+\n', udiff)[:0:-1]  # reversed order

    line_numbers = re.findall('@@\s[+-]\d+,\d+ \+(\d+)', udiff)
    line_numbers = list(map(int, line_numbers))[::-1]

    # Note: these were reversed as can modify number of lines
    for c, start in zip(chunks, line_numbers):
        ilines = enumerate((line for line in c.splitlines()
                            if not line.startswith('-')),
                           start=start)
        added_lines = [i for i, line in ilines if line.startswith('+')]
        if added_lines:
            yield (added_lines[0], added_lines[-1])


def udiff_lines_fixed(u):
    """Count lines fixed (removed) in udiff."""
    # TODO maybe this should return + and - (and tweak printing in Radius)
    removed_changes = re.findall('\n\-', u)
    return len(removed_changes)


def get_diff(original, fixed, file_name,
             original_label='original', fixed_label='fixed'):
    """Return text of unified diff between original and fixed."""
    original, fixed = original.splitlines(True), fixed.splitlines(True)
    newline = '\n'

    from difflib import unified_diff
    diff = unified_diff(original, fixed,
                        os.path.join(file_name, original_label),
                        os.path.join(file_name, fixed_label),
                        lineterm=newline)
    text = ''
    for line in diff:
        text += line
        # Work around missing newline (http://bugs.python.org/issue2142).
        if not line.endswith(newline):
            text += newline + r'\ No newline at end of file' + newline
    return text


def print_diff(diff, color=True):
    """Pretty printing for a diff, if color then we use a simple color scheme
    (red for removed lines, green for added lines)."""
    import colorama

    if not diff:
        return

    if not color:
        colorama.init = lambda autoreset: None
        colorama.Fore.RED = ''
        colorama.Back.RED = ''
        colorama.Fore.GREEN = ''
        colorama.deinit = lambda: None

    colorama.init(autoreset=True)  # TODO use context_manager
    for line in diff.splitlines():
        if line.startswith('+') and not line.startswith('+++ '):
            # Note there shouldn't be trailing whitespace
            # but may be nice to generalise this
            print(colorama.Fore.GREEN + line)
        elif line.startswith('-') and not line.startswith('--- '):
            split_whitespace = re.split('(\s+)$', line)
            if len(split_whitespace) > 1:  # claim it must be 3
                line, trailing, _ = split_whitespace
            else:
                line, trailing = split_whitespace[0], ''
            print(colorama.Fore.RED + line, end='')
            # give trailing whitespace a RED background
            print(colorama.Back.RED + trailing)
        elif line == '\ No newline at end of file':
            # The assumption here is that there is now a new line...
            print(colorama.Fore.RED + line)
        else:
            print(line)
    colorama.deinit()



"""This module does the argument and config parsing, and contains the main
function (that is called when calling pep8radius from shell)."""

from __future__ import print_function

import os
import sys

from argparse import ArgumentParser
try:
    from ConfigParser import SafeConfigParser, NoSectionError
except ImportError:  # py3, pragma: no cover
    from configparser import SafeConfigParser, NoSectionError

from pep8radius.radius import Radius
from pep8radius.shell import CalledProcessError  # with 2.6 compat

__version__ = version = '0.9.0'


DEFAULT_IGNORE = 'E24'
DEFAULT_INDENT_SIZE = 4

if sys.platform == 'win32':  # pragma: no cover
    DEFAULT_CONFIG = os.path.expanduser(r'~\.pep8')
else:
    DEFAULT_CONFIG = os.path.join(os.getenv('XDG_CONFIG_HOME') or
                                  os.path.expanduser('~/.config'), 'pep8')
PROJECT_CONFIG = ('setup.cfg', 'tox.ini', '.pep8')


def main(args=None, vc=None, cwd=None, apply_config=False):
    """PEP8 clean only the parts of the files touched since the last commit, a
    previous commit or branch."""
    import signal

    try:  # pragma: no cover
        # Exit on broken pipe.
        signal.signal(signal.SIGPIPE, signal.SIG_DFL)
    except AttributeError:  # pragma: no cover
        # SIGPIPE is not available on Windows.
        pass

    try:
        if args is None:
            args = []

        try:
            # Note: argparse on py 2.6 you can't pass a set
            # TODO neater solution for this!
            args_set = set(args)
        except TypeError:
            args_set = args  # args is a Namespace
        if '--version' in args_set or getattr(args_set, 'version', 0):
            print(version)
            sys.exit(0)
        if '--list-fixes' in args_set or getattr(args_set, 'list_fixes', 0):
            from autopep8 import supported_fixes
            for code, description in sorted(supported_fixes()):
                print('{code} - {description}'.format(
                    code=code, description=description))
            sys.exit(0)

        try:
            try:
                args = parse_args(args, apply_config=apply_config)
            except TypeError:
                pass  # args is already a Namespace (testing)
            r = Radius(rev=args.rev, options=args, vc=vc, cwd=cwd)
        except NotImplementedError as e:  # pragma: no cover
            print(e)
            sys.exit(1)
        except CalledProcessError as c:  # pragma: no cover
            # cut off usage and exit
            output = c.output.splitlines()[0]
            print(output)
            sys.exit(c.returncode)
        r.fix()

    except KeyboardInterrupt:  # pragma: no cover
        return 1


def create_parser():
    """Create the parser for the pep8radius CLI."""
    description = ("PEP8 clean only the parts of the files which you have "
                   "touched since the last commit, previous commit or "
                   "branch.")
    epilog = ("Run before you commit, against a previous commit or "
              "branch before merging.")
    parser = ArgumentParser(description=description,
                            epilog=epilog,
                            prog='pep8radius')

    parser.add_argument('rev',
                        help='commit or name of branch to compare against',
                        nargs='?')
    parser.add_argument('--version',
                        help='print version number and exit',
                        action='store_true')

    parser.add_argument('--config',
                        default='',
                        help='path to pep8 config file; '
                        "don't pass anything and global and local "
                        'config files will be used; pass False or '
                        'a non-existent file to use defaults')
    parser.add_argument('-d', '--diff', action='store_true', dest='diff',
                        help='print the diff of fixed source vs original')
    parser.add_argument('-i', '--in-place', action='store_true',
                        help="make the fixes in place; modify the files")

    parser.add_argument('--no-color', action='store_true',
                        help='do not print diffs in color '
                             '(default is to use color)')
    parser.add_argument('-v', '--verbose', action='count', dest='verbose',
                        default=0,
                        help='print verbose messages; '
                        'multiple -v result in more verbose messages '
                        '(one less -v is passed to autopep8)')

    ap = parser.add_argument_group('pep8', 'Pep8 options to pass to autopep8.')
    ap.add_argument('-p', '--pep8-passes', metavar='n',
                    default=-1, type=int,
                    help='maximum number of additional pep8 passes '
                    '(default: infinite)')
    ap.add_argument('-a', '--aggressive', action='count', default=0,
                    help='enable non-whitespace changes; '
                    'multiple -a result in more aggressive changes')
    ap.add_argument('--experimental', action='store_true',
                    help='enable experimental fixes')
    ap.add_argument('--exclude', metavar='globs',
                    help='exclude file/directory names that match these '
                    'comma-separated globs')
    ap.add_argument('--list-fixes', action='store_true',
                    help='list codes for fixes and exit; '
                    'used by --ignore and --select')
    ap.add_argument('--ignore', metavar='errors', default='',
                    help='do not fix these errors/warnings '
                    '(default: {0})'.format(DEFAULT_IGNORE))
    ap.add_argument('--select', metavar='errors', default='',
                    help='fix only these errors/warnings (e.g. E4,W)')
    ap.add_argument('--max-line-length', metavar='n', default=79, type=int,
                    help='set maximum allowed line length '
                    '(default: %(default)s)')
    ap.add_argument('--indent-size', default=DEFAULT_INDENT_SIZE,
                    type=int, metavar='n',
                    help='number of spaces per indent level '
                    '(default %(default)s)')

    df = parser.add_argument_group('docformatter',
                                   'Fix docstrings for PEP257.')
    df.add_argument('-f', '--docformatter', action='store_true',
                    help='Use docformatter')
    df.add_argument('--no-blank', dest='post_description_blank',
                    action='store_false',
                    help='Do not add blank line after description')
    df.add_argument('--pre-summary-newline',
                    action='store_true',
                    help='add a newline before the summary of a '
                    'multi-line docstring')
    df.add_argument('--force-wrap', action='store_true',
                    help='force descriptions to be wrapped even if it may '
                    'result in a mess')

    return parser


def parse_args(arguments=None, root=None, apply_config=False):
    """Parse the arguments from the CLI.

    If apply_config then we first look up and apply configs using
    apply_config_defaults.

    """
    if arguments is None:
        arguments = []

    parser = create_parser()
    if apply_config:
        apply_config_defaults(parser, arguments, root=root)
    args = parser.parse_args(arguments)

    # sanity check args (from autopep8)
    if args.max_line_length <= 0:  # pragma: no cover
        parser.error('--max-line-length must be greater than 0')

    if args.select:
        args.select = _split_comma_separated(args.select)

    if args.ignore:
        args.ignore = _split_comma_separated(args.ignore)
    elif not args.select and args.aggressive:
        # Enable everything by default if aggressive.
        args.select = ['E', 'W']
    else:
        args.ignore = _split_comma_separated(DEFAULT_IGNORE)

    if args.exclude:
        args.exclude = _split_comma_separated(args.exclude)
    else:
        args.exclude = []

    return args


def apply_config_defaults(parser, arguments, root):
    """Update the parser's defaults from either the arguments' config_arg or
    the config files given in config_files(root)."""
    if root is None:
        from pep8radius.vcs import VersionControl
        root = VersionControl.which().root_dir()

    config_file = config_arg(arguments)

    config = SafeConfigParser()
    config.read(config_file or config_files(root))
    try:
        defaults = dict((k.replace('-', '_'), v)
                        for k, v in config.items("pep8"))
        parser.set_defaults(**defaults)
    except NoSectionError:
        pass  # just do nothing, potentially this could raise ?
    return parser


def config_arg(arguments):
    """Get --config arg from arguments.
    """
    for arg in arguments:
        if arg.startswith('--config'):
            config_file = arg[9:]
            return os.path.expanduser(config_file)


def config_files(root, home='~'):
    """Returns a list of the global config files and local config files
    found in the root directory.

    Note as we pass this into config.read, where later config files are
    considered more important than preceding. That is later files override
    the settings of previous files.

    """
    home = os.path.expanduser('~')
    local_configs = [os.path.join(root, c) for c in PROJECT_CONFIG]
    global_configs = [DEFAULT_CONFIG] + [os.path.join(home, c)
                                         for c in PROJECT_CONFIG]
    return global_configs + local_configs


def _split_comma_separated(string):
    """Return a set of strings."""
    return set(filter(None, string.split(',')))


def _main(args=None, vc=None, cwd=None):  # pragma: no cover
    if args is None:
        args = sys.argv[1:]
    return main(args=args, vc=vc, cwd=cwd, apply_config=True)


if __name__ == "__main__":  # pragma: no cover
    _main()



from tests.test_radius import TestRadius, MixinTests
from tests.util import *
from tests.util_vcs import *


class TestRadiusNoVCS(TestCase):

    def __init__(self, *args, **kwargs):
        mk_temp_dirs()
        super(TestRadiusNoVCS, self).__init__(*args, **kwargs)

    def test_no_vc(self):

        # self.assertRaises(NotImplementedError, VersionControl.which)
        # This see the above repo, which is pep8radius and using git !
        raise SkipTest("TODO not sure how to test, as we're always in a repo!")

    def test_bad_vc(self):
        self.assertRaises(NotImplementedError,
                          lambda x: Radius(vc=x),
                          'made_up_vc')

    def test_using_vc(self, cwd=TEMP_DIR):
        # TODO dry this and move to TestRadius
        MixinGit.delete_repo()
        MixinHg.delete_repo()
        MixinBzr.delete_repo()

        self.assertFalse(using_hg(cwd=cwd))
        if MixinHg.init_vc(cwd=cwd):
            self.assertTrue(using_hg(cwd=cwd))
            self.assertTrue(isinstance(Radius(vc='hg', cwd=cwd).vc, Hg))

        self.assertFalse(using_bzr(cwd=cwd))
        if MixinBzr.init_vc(cwd=cwd):
            self.assertTrue(using_bzr(cwd=cwd))
            self.assertTrue(isinstance(Radius(vc='bzr', cwd=cwd).vc, Bzr))

        # git is seen before this, as the dir above is git!
        self.assertTrue(using_git(cwd=cwd))
        self.assertTrue(isinstance(Radius(vc='git').vc, Git))


class TestRadiusGit(TestRadius, MixinGit, MixinTests):
    vc = 'git'


class TestRadiusHg(TestRadius, MixinHg, MixinTests):
    vc = 'hg'


class TestRadiusBzr(TestRadius, MixinBzr, MixinTests):
    vc = 'bzr'



import autopep8
from tests.util import *


class TestMain(TestCase):

    def test_autopep8_args(self):
        import autopep8

        args = ['hello.py']
        us = parse_args(args)
        them = autopep8.parse_args(args)

        self.assertEqual(us.select, them.select)
        self.assertEqual(us.ignore, them.ignore)

        args = ['hello.py', '--select=E1,W1', '--ignore=W601',
                '--max-line-length', '120']
        us = parse_args(args)
        them = autopep8.parse_args(args)
        self.assertEqual(us.select, them.select)
        self.assertEqual(us.ignore, them.ignore)
        self.assertEqual(us.max_line_length, them.max_line_length)

        args = ['hello.py', '--aggressive', '-v']
        us = parse_args(args)
        them = autopep8.parse_args(args)
        self.assertEqual(us.aggressive, them.aggressive)

    def test_version_number(self):
        version_ = pep8radius_main(['--version'])
        self.assertEqual(version_, version)

    def test_list_fixes(self):
        fixes = pep8radius_main(['--list-fixes'])
        afixes = shell_out(['autopep8', '--list-fixes'])
        self.assertEqual(fixes, afixes)

    def test_config(self):
        cfg = os.path.join(TEMP_DIR, '.pep8')
        remove(cfg)

        args_before = parse_args(apply_config=False)
        self.assertNotIn('E999', args_before.ignore)

        with open(cfg, mode='w') as f:
            f.write("[pep8]\nignore=E999")
        args_after = parse_args(['--config=%s' % cfg], apply_config=True)
        self.assertIn('E999', args_after.ignore)
        args_after = parse_args(['--config=False'], apply_config=True)
        self.assertNotIn('E999', args_after.ignore)

        remove(cfg)

    def test_help(self):
        self.check_help()

    def test_help_outside_project(self):
        home = os.path.expanduser('~')
        try:
            VersionControl.which(cwd=home)
            raise SkipTest("Home directory is under version control ??")
        except NotImplementedError:
            pass
        self.check_help(cwd=home)

    def check_help(self, cwd=TEMP_DIR):
        help_message = pep8radius_main(['--help'], cwd=cwd)
        self.assertIn("--no-color", help_message)
        self.assertIn("PEP8 clean only the parts of the files", help_message)

if __name__ == '__main__':
    test_main()



"""Some additional vcs funtionality not provided in pep8radius."""

from tests.util import *


class MixinVcs(object):

    @classmethod
    def save_and_commit(cls, contents, f):
        save(contents, f)
        return cls.successfully_commit_files([f])

    @classmethod
    def init_vc(cls, cwd=TEMP_DIR):
        cls.delete_repo(cwd=cwd)
        success = cls.create_repo(cwd=cwd)
        committed = cls.save_and_commit('a=1;', 'a.py')
        return success and committed


class MixinGit(MixinVcs):

    @staticmethod
    def delete_repo(cwd=TEMP_DIR):
        temp_path = os.path.join(cwd, '.git')
        remove_dir(temp_path)

    @staticmethod
    def create_repo(cwd=TEMP_DIR):
        try:
            shell_out(["git", "init"], cwd=cwd)
            return True
        except (OSError, CalledProcessError):
            return False

    @staticmethod
    def successfully_commit_files(file_names,
                                  commit="initial_commit",
                                  cwd=TEMP_DIR):
        try:
            shell_out(["git", "add"] + file_names, cwd=cwd)
            shell_out(["git", "commit", "-m", commit], cwd=cwd)
            with from_dir(cwd):
                return Git().current_branch()
        except (OSError, CalledProcessError):
            return False

    @staticmethod
    def checkout(branch, create=False, cwd=TEMP_DIR):
        if create:
            shell_out(["git", "checkout", '-b', branch], cwd=cwd)
        else:
            shell_out(["git", "checkout", branch], cwd=cwd)


class MixinHg(MixinVcs):

    @staticmethod
    def delete_repo(cwd=TEMP_DIR):
        remove_dir(os.path.join(cwd, '.hg'))

    @staticmethod
    def create_repo(cwd=TEMP_DIR):
        try:
            shell_out(["hg", "init"], cwd=cwd)
            return True
        except (OSError, CalledProcessError):
            return False

    @staticmethod
    def successfully_commit_files(file_names,
                                  commit="initial_commit",
                                  cwd=TEMP_DIR):
        try:
            shell_out(["hg", "add"] + file_names, cwd=cwd)
            shell_out(["hg", "commit", "-m", commit], cwd=cwd)
            with from_dir(cwd):
                return Hg().current_branch()
        except (OSError, CalledProcessError):
            return False

    @staticmethod
    def checkout(branch, create=False, cwd=TEMP_DIR):
        if create:
            shell_out(["hg", "branch", branch], cwd=cwd)
        else:
            shell_out(["hg", "update", "--check", branch], cwd=cwd)


class MixinBzr(MixinVcs):

    @staticmethod
    def delete_repo(cwd=TEMP_DIR):
        remove_dir(os.path.join(cwd, '.bzr'))

    @staticmethod
    def create_repo(cwd=TEMP_DIR):
        try:
            shell_out(["bzr", "init"], cwd=cwd)
            return True
        except (OSError, CalledProcessError):
            return False

    @staticmethod
    def successfully_commit_files(file_names,
                                  commit="initial_commit",
                                  cwd=TEMP_DIR):
        try:
            shell_out(["bzr", "add"] + file_names, cwd=cwd)
            shell_out(["bzr", "commit", "-m", commit], cwd=cwd)
            with from_dir(cwd):
                return Bzr().current_branch()
        except (OSError, CalledProcessError):
            return False

    @staticmethod
    def checkout(branch, create=False, cwd=TEMP_DIR):
        create = ['--create-branch'] if create else []
        shell_out(["bzr", "switch", branch] + create, cwd=cwd)



"""This module defines some helper methods shell_out and
shell_out_ignore_exitcode, which are lightweight wrappers around subprocess'
check_output function.

Note: We also monkey-patch subprocess for python 2.6 to
give feature parity with later versions.

"""


try:
    from subprocess import STDOUT, check_output, CalledProcessError
except ImportError:  # pragma: no cover
    # python 2.6 doesn't include check_output
    # monkey patch it in!
    import subprocess
    STDOUT = subprocess.STDOUT

    def check_output(*popenargs, **kwargs):
        if 'stdout' in kwargs:  # pragma: no cover
            raise ValueError('stdout argument not allowed, '
                             'it will be overridden.')
        process = subprocess.Popen(stdout=subprocess.PIPE,
                                   *popenargs, **kwargs)
        output, _ = process.communicate()
        retcode = process.poll()
        if retcode:
            cmd = kwargs.get("args")
            if cmd is None:
                cmd = popenargs[0]
            raise subprocess.CalledProcessError(retcode, cmd,
                                                output=output)
        return output
    subprocess.check_output = check_output

    # overwrite CalledProcessError due to `output`
    # keyword not being available (in 2.6)
    class CalledProcessError(Exception):

        def __init__(self, returncode, cmd, output=None):
            self.returncode = returncode
            self.cmd = cmd
            self.output = output

        def __str__(self):
            return "Command '%s' returned non-zero exit status %d" % (
                self.cmd, self.returncode)
    subprocess.CalledProcessError = CalledProcessError


def shell_out(cmd, stderr=STDOUT, cwd=None):
    """Friendlier version of check_output."""
    if cwd is None:
        from os import getcwd
        cwd = getcwd()  # TODO do I need to normalize this on Windows

    out = check_output(cmd, cwd=cwd, stderr=stderr, universal_newlines=True)
    return _clean_output(out)


def shell_out_ignore_exitcode(cmd, stderr=STDOUT, cwd=None):
    """Same as shell_out but doesn't raise if the cmd exits badly."""
    try:
        return shell_out(cmd, stderr=stderr, cwd=cwd)
    except CalledProcessError as c:
        return _clean_output(c.output)


def _clean_output(out):
    try:
        out = out.decode('utf-8')
    except AttributeError:  # python3, pragma: no cover
        pass
    return out.strip()



"""This module defines the Radius class, which is where the "meat" of the
pep8radius machinery is done. The important methods are fix, fix_file and
fix_line_range.

The vc attribute is a subclass of VersionControl defined in the vcs
module (provides helper methods for the different vcs e.g. git).

"""

from __future__ import print_function

from sys import version_info

from pep8radius.diff import get_diff, udiff_lines_fixed, print_diff
from pep8radius.vcs import VersionControl


if version_info[0] > 2:  # py3, pragma: no cover
    basestring = str


class Radius(object):

    """PEP8 clean only the parts of the files touched since the last commit, a
    previous commit or branch."""

    def __init__(self, rev=None, options=None, vc=None, cwd=None):
        if vc is None:
            vc = VersionControl.which()
        elif isinstance(vc, basestring):
            vc = VersionControl.from_string(vc)
        else:
            assert(issubclass(vc, VersionControl))
        self.vc = vc(cwd=cwd)

        # pep8radius specific options
        self.rev = self.vc.branch_point(rev)
        from pep8radius.main import parse_args
        self.options = options if options else parse_args([''])
        self.verbose = self.options.verbose
        self.in_place = self.options.in_place
        self.diff = self.options.diff
        self.color = not self.options.no_color

        # autopep8 specific options
        self.options.verbose = max(0, self.options.verbose - 1)
        self.options.in_place = False
        self.options.diff = False

        # Note: This may raise a CalledProcessError, if it does it means
        # that there's been an error with the version control command.
        self.filenames_diff = self.vc.get_filenames_diff(self)

    def fix(self):
        """Runs fix_file on each modified file.

        - Prints progress and diff depending on options.

        """
        n = len(self.filenames_diff)

        self.p('Applying autopep8 to touched lines in %s file(s).' % n)

        total_lines_changed = 0
        pep8_diffs = []
        for i, file_name in enumerate(self.filenames_diff, start=1):
            self.p('%s/%s: %s: ' % (i, n, file_name), end='')
            self.p('', min_=2)

            p_diff = self.fix_file(file_name)
            lines_changed = udiff_lines_fixed(p_diff) if p_diff else 0
            total_lines_changed += lines_changed

            if p_diff and self.diff:
                pep8_diffs.append(p_diff)

        if self.in_place:
            self.p('pep8radius fixed %s lines in %s files.'
                   % (total_lines_changed, n))
        else:
            self.p('pep8radius would fix %s lines in %s files.'
                   % (total_lines_changed, n))

        if self.diff:
            for diff in pep8_diffs:
                print_diff(diff, color=self.color)

    def fix_file(self, file_name):
        """Apply autopep8 to the diff lines of a file.

        - Returns the diff between original and fixed file.
        - If self.in_place then this writes the the fixed code the file_name.
        - Prints dots to show progress depending on options.

        """
        import codecs
        try:
            with codecs.open(file_name, 'r', encoding='utf-8') as f:
                original = f.read()
        except IOError:
            # file has been removed
            return ''

        # We hope that a CalledProcessError would have already raised
        # during the init if it were going to raise here.
        modified_lines_descending = self.vc.modified_lines(self, file_name)

        # Apply line fixes "up" the file (i.e. in reverse) so that
        # fixes do not affect changes we're yet to make.
        partial = original
        for start, end in modified_lines_descending:
            partial = self.fix_line_range(partial, start, end)
            self.p('.', end='', max_=1)
        self.p('', max_=1)
        fixed = partial

        if self.in_place:
            with codecs.open(file_name, 'w', encoding='utf-8') as f:
                f.write(fixed)
        return get_diff(original, fixed, file_name)

    def fix_line_range(self, f, start, end):
        """Apply autopep8 between start and end of file f."""
        # not sure on behaviour if outside range (indexing starts at 1)
        start = max(start, 1)

        self.options.line_range = [start, end]
        from autopep8 import fix_code
        fixed = fix_code(f,   self.options)

        if self.options.docformatter:
            from docformatter import format_code
            fixed = format_code(
                fixed,
                summary_wrap_length=self.options.max_line_length - 1,
                description_wrap_length=(self.options.max_line_length
                                         - 2 * self.options.indent_size),
                pre_summary_newline=self.options.pre_summary_newline,
                post_description_blank=self.options.post_description_blank,
                force_wrap=self.options.force_wrap,
                line_range=[start, end])

        return fixed

    def p(self, something_to_print, end=None, min_=1, max_=99):
        """Print if self.verbose is within min_ and max_."""
        if min_ <= self.verbose <= max_:
            import sys
            print(something_to_print, end=end)
            sys.stdout.flush()



"""Everythin' you'll ever need in the tests..."""

from __future__ import with_statement

from contextlib import contextmanager
import os
from shutil import rmtree
import sys

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

if sys.version_info < (2, 7):
    from unittest2 import main as test_main, SkipTest, TestCase
else:
    from unittest import main as test_main, SkipTest, TestCase


ROOT_DIR = os.path.split(os.path.abspath(os.path.dirname(__file__)))[0]
sys.path.insert(0, ROOT_DIR)
from pep8radius import (Radius,
                        shell_out,
                        parse_args,
                        version)
from pep8radius.diff import modified_lines_from_udiff, get_diff
from pep8radius.shell import CalledProcessError
from pep8radius.vcs import (VersionControl, Git, Bzr, Hg,
                            using_git, using_hg, using_bzr)


PEP8RADIUS = os.path.join(ROOT_DIR, 'pep8radius', '__init__.py')
TEST_DIR = os.path.abspath(os.path.dirname(__file__))
TEMP_DIR = os.path.join(TEST_DIR, 'temp')
SUBTEMP_DIR = os.path.join(TEMP_DIR, 'subtemp')


@contextmanager
def captured_output():
    new_out, new_err = StringIO(), StringIO()
    old_out, old_err = sys.stdout, sys.stderr
    try:
        sys.stdout, sys.stderr = new_out, new_err
        yield sys.stdout, sys.stderr
    finally:
        sys.stdout, sys.stderr = old_out, old_err


@contextmanager
def from_dir(cwd):
    curdir = os.getcwd()
    try:
        os.chdir(cwd)
        yield
    finally:
        os.chdir(curdir)


def pep8radius_main(args, vc=None, cwd=TEMP_DIR, apply_config=False):
    with from_dir(cwd):
        with captured_output() as (out, err):
            try:
                from pep8radius.main import main
                main(args, vc=vc, apply_config=apply_config, cwd=cwd)
            except SystemExit:
                pass
        return out.getvalue().strip()


def mk_temp_dirs():
    try:
        os.mkdir(TEMP_DIR)
    except OSError:
        pass

    try:
        os.mkdir(SUBTEMP_DIR)
    except OSError:
        pass


def save(contents, f, cwd=TEMP_DIR):
    with from_dir(cwd):
        with open(f, 'w') as f1:
            f1.write(contents)


def remove(filename):
    """Delete file filename and don't raise if missing."""
    try:
        os.remove(filename)
    except OSError:
        pass


def remove_dir(directory):
    """Delete directory and all contained files and subdirectories.

    Note: this is the same as the shutil's rmtree function but overcomes
    a reported issue on Windows*.

    *Due to permissions (being able to create but not remove files).

    """
    try:
        rmtree(directory)
    except OSError as e:  # pragma: no cover
        # see http://stackoverflow.com/questions/1213706/
        # and http://stackoverflow.com/questions/7228296/
        from errno import EACCES
        if e.errno == EACCES:
            from stat import S_IWRITE
            for dirpath, dirnames, filenames in os.walk(directory):
                for filename in filenames:
                    os.chmod(os.path.join(dirpath, filename), S_IWRITE)
            rmtree(directory)


def get_diff_many(modified, expected, files):
    return ''.join(get_diff(*mef)
                   for mef in zip(modified, expected, files))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from ast import parse
import os
from setuptools import setup
from sys import version_info


NAME = 'pep8radius'  # 'Better-Than-You-Found-It'


def version():
    """Return version string."""
    with open(os.path.join(os.path.abspath(os.path.dirname(__file__)),
                           'pep8radius',
                           'main.py')) as input_file:
        for line in input_file:
            if line.startswith('__version__'):
                return parse(line).body[0].value.s


def readme():
    try:
        import pypandoc
        return pypandoc.convert('README.md', 'rst', format='md')
    except (IOError, ImportError):
        with open('README.md') as f:
            return f.read()

INSTALL_REQUIRES = (
    ['autopep8 >= 1.0.4'] +
    (['argparse'] if version_info < (2, 7) else []) +
    ['colorama'] +
    ['docformatter >= 0.7']
)

setup(
    name=NAME,
    version=version(),
    description="PEP8 clean only the parts of the files which you have touched"
                " since the last commit, previous commit or branch.",
    long_description=readme(),
    license='MIT License',
    author='Andy Hayden',
    author_email='andyhayden1@gmail.com',
    url='https://github.com/hayd/pep8radius',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: Quality Assurance',
        'Topic :: Software Development :: Version Control',
    ],
    keywords='automation, pep8, format, autopep8, git, hg, mercurial, bzr',
    install_requires=INSTALL_REQUIRES,
    packages=['pep8radius'],
    test_suite='tests',
    zip_safe=False,
    entry_points={'console_scripts': ['btyfi = pep8radius.main:_main',
                                      'pep8radius = pep8radius.main:_main']},
)
